module Scrivito

# The CMS widget class. Widget classes let you create widget types for the individual kinds
# of content editors want to place on pages or into widgets. All widget classes are based on
# subclasses of {Scrivito::BasicWidget Widget}. As with CMS objects, the content of a widget
# is stored in its attributes.
# @see https://scrivito.com/page-with-widgets Creating a page type with widgets
# @api public
class BasicWidget
  extend AttributeContent::ClassMethods
  extend ::ActiveSupport::DescendantsTracker
  include AttributeContent

  def self.type_computer
    @_type_computer ||= TypeComputer.new(Scrivito::BasicWidget, ::Widget)
  end

  def self.reset_type_computer!
    @_type_computer = nil
  end

  #
  # @api public
  # This method can be overridden in subclasses to control to which pages or
  # widgets the given widget class can be added. This method can be used
  # to ensure that only a special type of widget can be added to a specific
  # container. An example for this is a +TabGroupWidget+ that should
  # contain widgets of the +TabWidget+ type only, and, vice versa, a +TabWidget+
  # should only be contained in a +TabGroupWidget+. A value of +NilClass+ means that
  # no additional restrictions are applied.
  #
  # This method only further restricts the list of valid classes defined
  # by means of {AttributeContent#valid_widget_classes_for}.
  #
  # @return [NilClass, Array<Class>]
  #
  # @example
  #   class TabGroupWidget < Widget
  #     def valid_widget_classes_for(field_name)
  #       [TabWidget]
  #     end
  #   end
  #
  #   class TabWidget < Widget
  #     def self.valid_container_classes
  #       [TabGroupWidget]
  #     end
  #   end
  #
  def self.valid_container_classes
  end

  def self.valid_inside_container?(container_class)
    computed_classes = assert_classes(valid_container_classes, '.valid_container_classes')

    computed_classes.nil? || computed_classes.include?(container_class)
  end

  # @api beta
  def self.embeds(attribute_name)
    @embedding_attribute = attribute_name
    nil
  end

  # @api beta
  def self.embeds?
    !!@embedding_attribute || !!superclass.try(:embeds?)
  end

  # @api beta
  def self.embedding_attribute
    raise ScrivitoError, "#{self} has no embedding attribute" unless embeds?

    if @embedding_attribute
      unless attribute_definition = attribute_definitions[@embedding_attribute]
        raise ScrivitoError, %{Unknown attribute: "#{@embedding_attribute}"}
      end

      unless attribute_definition.reference?
        raise ScrivitoError, %{Embedding attribute "#{@embedding_attribute}" must be a reference}
      end

      attribute_definition
    else
      superclass.embedding_attribute
    end
  end

  def self.as_json
    super.merge(
      embeds: embeds?,
      embeddingAttribute: @embedding_attribute,
    )
  end

  attr_accessor :container, :container_attribute_name

  attr_reader :attributes_to_be_saved

  delegate :widget_from_pool, to: :obj

  # Create a new Widget.
  # The new Widget must be stored inside a container (i.e. an Obj or another Widget)
  # before it can be used.
  #
  # See {Scrivito::BasicObj.create Obj.create} for a detailed overview of how to set attributes.
  # @api public
  # @param [Hash] attributes
  #
  # @example Create a widget using a subclass:
  #   # you can create widgets by explicitly providing the '_obj_class' attribute
  #   new_widget = Widget.new(_obj_class: "MyWidget")
  #   # but it is better to simply use the constructor of a subclass
  #   new_widget = MyWidget.new
  #
  # @example Create a widget and store it inside an Obj:
  #   # create the widget
  #   new_widget = Widget.new(_obj_class: "MyWidget")
  #   # store it inside an obj
  #   my_obj(my_widget_field: [new_widget])
  #
  def initialize(attributes = {})
    @attributes_to_be_saved = self.class.prepare_attributes_for_instantiation(attributes)
    @attribute_cache = {}
  end

  #
  # Creates a new {Scrivito::BasicWidget Widget}. The defaults set via
  # {Scrivito::AttributeContent::ClassMethods#default_for Widget.default_for}
  # are taken into account.
  #
  # @api public
  #
  # @note Note that creating a widget does not implicitly persist it. The widget is only persisted
  #   if its creation is part of the creation of the object containing it, or as the object or
  #   widget containing it is updated.
  #
  # @param [Hash] attributes for the new widget
  # @param [Hash] context in which the object creating should happen
  # @option context [Scrivito::User] :scrivito_user current visitor
  # @option context [Scrivito::BasicObj] :current_page current page
  # @return [Widget] the newly created {Scrivito::BasicWidget Widget}
  #
  # @raise [ArgumentError] if called on +Widget+ without +_obj_class+ attribute.
  #
  # @see Scrivito::AttributeContent::ClassMethods#default_for
  #
  def self.new(attributes = {}, context = {})
    if obj_class = extract_obj_class_from_attributes(attributes)
      obj_class.new(attributes, context)
    else
      super(build_attributes_with_defaults(attributes, context))
    end
  end

  def revision
    obj.revision
  end

  # Updates the attributes of this {Scrivito::BasicWidget Widget}.
  # See {Scrivito::BasicObj.create Obj.create} for a detailed overview of how to set attributes.
  # @example Set the title of a specific widget contained in the 'main_content' attribute of an Obj:
  #   obj.main_content[4].update(title: "New Title")
  #
  # @example Add a widget to the 'main_content' widgetlist attribute:
  #   widgets = obj.main_content << MyWidget.new(title: "Fresh")
  #   obj.update(main_content: widgets)
  # @api public
  # @param [Hash] attributes
  def update(attributes)
    obj.update(_widget_pool: { self => attributes })
    reload
  end

  # Destroys the {Scrivito::BasicWidget Widget} in the current {Workspace}.
  # @api public
  def destroy
    new_widget_list = container[container_attribute_name] - [self]
    container.update(container_attribute_name => new_widget_list)
  end

  #
  # Create a copy of a {Scrivito::BasicWidget Widget}.
  #
  # @api public
  #
  # The copy will have all the attributes of the original widget including its widgets.
  # Its attributes can be accessed only after it has been stored in a +widgetlist+ field of an
  # {Scrivito::BasicObj Obj}, since initially the copy is not stored in such a field.
  #
  # @example Duplicate the first widget contained in the +my_widget+ field.
  #   obj.update(my_widgets: obj.my_widgets.push(obj.my_widgets.first.copy))
  #
  # @return [Scrivito::BasicWidget] a copy of this widget
  def copy
    attributes = {}
    attribute_definitions.each do |attribute_definition|
      attribute_name = attribute_definition.name
      attribute_value = read_attribute(attribute_name)
      attribute_value = attribute_value.map(&:copy) if attribute_definition.widgetlist?
      attributes[attribute_name] = attribute_value
    end
    self.class.new(attributes)
  end

  def clone
    raise "The method `clone' was removed. Please use `copy' instead"
  end

  def id
    if @id
      @id
    else
      raise_not_persisted_error
    end
  end

  def persisted?
    @attributes_to_be_saved.nil?
  end

  def attach_to_obj(an_obj, widget_id = compute_widget_id(an_obj))
    @obj = an_obj
    @id = widget_id

    attributes_to_be_saved
  end

  def ==(other)
    eql?(other)
  end

  def eql?(other)
    other.respond_to?(:equality_identifier, true) &&
      equality_identifier.eql?(other.equality_identifier)
  end

  def hash
    unless persisted?
      # An unbound widget will likely be attached soon, which mutates its #hash value. Evil.
      raise ScrivitoError, 'a new widget is not allowed to be used as a hash key'
    end

    equality_identifier.hash
  end

  #
  # Reverts all changes made to the +Widget+ in the current workspace.
  #
  # @api public
  #
  # @note This method does not support +new+ +Widget+s.
  #   Please use {Scrivito::BasicWidget#destroy Widget#destroy} to destroy them.
  # @note This method does not support +deleted+ +Widget+s.
  #   Please use {Scrivito::BasicObj#revert Obj#revert} to restore them.
  #
  # @raise [ScrivitoError] If the current workspace is +published+.
  # @raise [ScrivitoError] If the +Widget+ is +new+.
  # @raise [ScrivitoError] If the +Widget+ is +deleted+.
  #
  def revert
    workspace.assert_revertable

    case modification
    when Modification::UNMODIFIED
      # do nothing
    when Modification::EDITED
      previous_obj_content =
          CmsRestApi.get("revisions/#{workspace.base_revision_id}/objs/#{obj.id}")
      previous_widget_content = previous_obj_content["_widget_pool"]["#{id}"]
      previous_widget_content = self.class.reset_blank_attributes(previous_widget_content)

      previous_widget_content.delete_if do |attribute_name, _|
        type_of_attribute(attribute_name) == 'widgetlist'
      end
      CmsRestApi.put("workspaces/#{workspace.id}/objs/#{obj.id}",
          { obj: {_widget_pool: {id => previous_widget_content}} })
      reload
    else
      raise ScrivitoError, "cannot revert changes, since widget is #{modification}"
    end
  end

  def in_revision(revision)
    obj_in_revision = obj.in_revision(revision)
    obj_in_revision && obj_in_revision.widget_from_pool(id)
  end

  def new?(revision=workspace.base_revision)
    return false unless revision
    obj.new?(revision) || cms_data_for_revision(revision).nil?
  end

  def deleted?(revision=workspace.base_revision)
    return false unless revision
    obj.deleted?(revision)
  end

  def modification(revision=workspace.base_revision)
    return Modification::UNMODIFIED unless revision

    if deleted?(revision)
      Modification::DELETED
    elsif new?(revision)
      Modification::NEW
    else
      if data_from_cms == cms_data_for_revision(revision)
        Modification::UNMODIFIED
      else
        Modification::EDITED
      end
    end
  end

  # Returns the entity ({Scrivito::BasicObj} or {Scrivito::BasicWidget})
  # that references this widget.
  # @api public
  #
  # @return [Scrivito::AttributeContent] the entity that references this widget.
  def container
    @container || container_and_attribute_name.first
  end

  # Returns the name of the +widgetlist+ attribute that references this widget.
  # @return [String] the name of the attribute that references this widget.
  # @api public
  def container_attribute_name
    @container_attribute_name || container_and_attribute_name.second
  end

  # Returns the name of the 'widgetlist' field that references this widget.
  # @api public
  # @return [String] the name of the attribute that references this widget.
  # @deprecated
  def container_field_name
    Scrivito::Deprecation.warn('Scrivito::BasicWidget#container_field_name is deprecated'\
      ' and will be removed in a future version.'\
      ' Please use Scrivito::BasicWidget#container_attribute_name instead.')
    container_attribute_name
  end

  def has_attribute?(key)
    key.to_s == '_obj_class' || super
  end

  def inspect
    if @id
      if @obj
        "<#{self.class} id=\"#{id}\" obj.id=\"#{obj.id}\">"
      else
        "<#{self.class} id=\"#{id}\">"
      end
    else
      "<#{self.class}>"
    end
  end

  # Returns the {Scrivito::BasicObj Obj} to which this widget belongs.
  # @api public
  #
  # @return [Scrivito::BasicObj] - the Obj this widget belongs to.
  def obj
    if @obj
      @obj
    else
      raise_not_persisted_error
    end
  end

  def notify_persisted(obj)
    @obj = obj
    @attributes_to_be_saved = nil
    reload_data
  end

  def reload
    obj.reload
    reload_data
  end

  def reload_data
    update_data -> { obj.widget_data_from_pool(id) }
  end

  # @api public
  # This method determines the description to be shown as widget tooltips. By default,
  # it uses the value of the +Widget.description_for_editor+ class method.
  #
  # This method can be overridden to customize the description displayed to the editor.
  #
  # @return [String] description for the editor
  def description_for_editor
    self.class.description_for_editor
  end

  def data_from_cms
    if persisted?
      super
    else
      raise_not_persisted_error
    end
  end

  def contained_widgets
    referenced = referenced_widgets
    referenced + referenced.map { |widget| widget.contained_widgets }.flatten
  end

  protected

  def equality_identifier
    [id, obj.id]
  end

  private

  def workspace
    if revision.workspace
      revision.workspace
    else
      raise ScrivitoError, "No workspace set for the obj of this widget"
    end
  end

  def raise_not_persisted_error
    raise ScrivitoError.new('Can not access a new widget before it has been saved')
  end

  def cms_data_for_revision(revision)
    obj.widget_data_for_revision(id, revision)
  end

  def container_and_attribute_name
    @container_and_attribute_name ||= obj.find_container_and_attribute_name_for_widget(id)
  end

  def has_system_attribute?(attribute_name)
    attribute_name == '_obj_class'
  end

  alias_method :has_public_system_attribute?, :has_system_attribute?

  def type_of_system_attribute(attribute_name)
    'string' if attribute_name == 'obj_class'
  end

  def value_of_system_attribute(attribute_name)
    data_from_cms.value_of(attribute_name)
  end

  def compute_widget_id(obj)
    @attributes_to_be_saved.delete('_id') || generate_widget_id(obj)
  end

  def generate_widget_id(obj)
    obj.present? ? obj.generate_widget_pool_id : BasicObj.generate_widget_pool_id
  end
end

end
